"""A Collection of derivatives for the activation functions"""

import numpy as np
import cupy as cp

from shamboflow import IS_CUDA
from shamboflow.engine import activations

def d_sigmoid(x : np.ndarray, **kwargs) -> np.ndarray :
    """The partial derivative of sigmoid function
    
    If sigmoid is g(x)
    ``g'(x) = g(x)(1 - g(x))`` 
    """

    if IS_CUDA :
        res_sig = cp.asarray(activations.sigmoid(x))
        res_d_sig = cp.multiply(res_sig, cp.subtract(1, res_sig))
        return cp.asnumpy(res_d_sig)
    
    res_sig = activations.sigmoid(x)
    res_d_sig = np.multiply(res_sig, np.subtract(1, res_sig))
    return res_d_sig

def d_tanh(x : np.ndarray, **kwargs) -> np.ndarray :
    """The partial derivative for hyperbolic tangent
    
    If tanh is g(x)
    ``g'(x) = 1 - g(x)^2``
    
    """

    if IS_CUDA :
        res_tanh = cp.asarray(activations.tanh(x))
        res_d_tanh = cp.subtract(1, cp.power(res_tanh, 2))
        return cp.asnumpy(res_d_tanh)
    
    res_tanh = activations.tanh(x)
    res_d_tanh = np.subtract(1, np.power(res_tanh, 2))
    return res_d_tanh

def d_relu(x : np.ndarray, **kwargs) -> np.ndarray :
    """Partial derivative for relu function

    ``Formula: 0 if x < 0 else 1``
    
    """

    if IS_CUDA :
        x_gpu = cp.asarray(x)
        res = cp.where(x_gpu < 0, 0, 1)
        return cp.asnumpy(res)
    
    return np.where(x < 0, 0, 1)

def d_leakyrelu(x : np.ndarray, leakyrelu_slope : float, **kwargs) -> np.ndarray :
    """Partial derivative for leaky relu function
    
    ``Formula: slope if x < 0 else 1``
    """

    if IS_CUDA :
        x_gpu = cp.asarray(x)
        res = np.where(x_gpu < 0, leakyrelu_slope, 1)
        return cp.asnumpy(res)
    
    return np.where(x < 0, leakyrelu_slope, 1)

def d_softmax(x : np.ndarray, **kwargs) -> np.ndarray :
    """Not implemented yet"""
    pass


def get(func : str) :
    """Helper function to get an activation function derivative

    Return the appropriate activation function
    depending on the given string

    Args
        func:
            Query string for the requested activation function derivative

    Returns
        Appropriate function
    """
    # TODO : handle wrong inputs
    if not isinstance(func, str) :
        ...

    func = func.strip().lower()

    if func == "sigmoid" :
        return d_sigmoid
    elif func == "tanh" :
        return d_tanh
    elif func == "relu" :
        return d_relu
    elif func == "leakyrelu" :
        return d_leakyrelu
    # elif func == "softmax" :
    #     return softmax


"""A small collection of loss functions"""

import numpy as np
import cupy as cp

from shamboflow import IS_CUDA

def MSE(pred_val : np.ndarray, obs_val : np.ndarray) -> float :
    """Mean Squared Error

    It is a loss/cost function to
    calculate error between the
    predicted data and the actual
    data.

    ``MSE = (1 / 2)(Sum (i = 1 to n)(y_i - y^_i) ^ 2)``

    It is the mean of all errors squared.
    It is multiplied with 1/2 instead of
    1/n to simplify the partial derivative.

    Args
        pred_val:
            The input vector to calculate the error for
        obs_val:
            The actual observed values
    
    Returns
        Overall error of the output

    """

    if IS_CUDA :
        obs_val = cp.asarray(obs_val)
        pred_val = cp.asarray(pred_val)
        sq_err = cp.power(cp.subtract(obs_val, pred_val), 2)
        mse_val = (1 / 2) * cp.sum(sq_err)

        return mse_val.item()

    sq_err = np.power(np.subtract(obs_val, pred_val), 2)
    mse_val = (1 / 2) * np.sum(sq_err)

    return mse_val.item()


def get(func : str) :
    """Helper function to get a loss function

    Return the appropriate loss function
    depending on the given string

    Args
        func:
            Query string for the requested loss function

    Returns
        Appropriate function
    """

    func = func.strip().lower()

    if func == "mean_squared_error" :
        return MSE


"""Common Activation functions"""

import numpy as np
import cupy as cp
from shamboflow import IS_CUDA

def sigmoid(x : np.ndarray, **kwargs) -> np.ndarray :
    """Sigmoid acitvation function
    
    Applies the sigmoid activation function to a value.
    It returns a value between 0 and 1 but never 0 and 1.
    It suffers from the vanishing gradient issue.

    ``sigmoid(x) = 1 / (1 + exp(-x))``

    Range: (0, 1)

    Args
        x:
            The input vector to apply activation function over.

    Returns
        Value after sigmoid is applied on x
    
    """

    if IS_CUDA :
        try :
            x_gpu = cp.asarray(x)
            res = (1 / (1 + cp.exp(-x_gpu)))
            return cp.asnumpy(res)
        except :
            pass

    return (1 / (1 + np.exp(-x)))

def tanh(x : np.ndarray, **kwargs) -> np.ndarray :
    """Hyperbolic tangent activation function

    ``tanh(x) = (exp(x) - exp(-x)) / (exp(x) + exp(-x))``

    Range: (-1, 1)

    Args
        x:
            The input vector to apply activation function over.

    Returns
        Value after tanh is applied on x
    
    """

    if IS_CUDA :
        try :
            x_gpu = cp.asarray(x)
            res = (cp.exp(x_gpu) - cp.exp(-x_gpu)) / (cp.exp(x_gpu) + cp.exp(-x_gpu))
            return cp.asnumpy(res)
        except :
            pass
    
    return ((np.exp(x) - np.exp(-x)) / (np.exp(x) + np.exp(-x)))

def relu(x : np.ndarray, **kwargs) -> np.ndarray :
    """ReLU activation function
    
    Applies the ReLU activation function on the input x.

    ReLU or Rectilinear Unit was devised to overcome
    the issue of vanishing gradient in sigmoid function. It resolves
    the issue by not having any upper limit. But that comes with
    its own problems.

    ``ReLU(x) = max(0, x)``

    Range: [0, inf)

    Args
        x:
            The input vector to apply activation function over.

    Returns
        Value after relu is applied on x
    """

    if IS_CUDA :
        try :
            x_gpu = cp.asarray(x)
            res = cp.fmax(0, x_gpu)
            return cp.asnumpy(res)
        except :
            pass

    return np.fmax(0, x)

def leakyrelu(x : np.ndarray, leakyrelu_slope : float, **kwargs) -> np.ndarray :
    """Leaky ReLU activation function
    
    This function was devised to address
    the issue of negative values as input
    to ReLU. ReLU would simply discard
    them, but leaky relu uses them as well.

    ``leakyReLU(x, slope) = if (x > 0) => x, if (x <= 0) => slope * x``

    Range: (-inf, inf)

    Args
        x:
            The input vector to apply activation function over.
    slope:
        slope of the line that provides output for negative inputs

    Returns
        Value after leakyrelu is applied on x
    """

    if IS_CUDA :
        try :
            x_gpu = cp.asarray(x)
            res = cp.where(x_gpu > 0, x_gpu, cp.multiply(leakyrelu_slope, x_gpu))
            return cp.asnumpy(res)
        except :
            pass

    return np.where(x > 0, x, np.multiply(leakyrelu_slope, x))

def softmax(x : np.ndarray, **kwargs) -> np.ndarray :
    """SoftMax activation function
    
    This function takes a vector as input
    and changes them to probabilities based
    on their values with respect to the
    vector itself.

    Range: (0, 1)

    Args
        x:
            The input vector to apply activation function over.

    Returns
        Value after softmax is applied on x

    WARNING: Not available at the moment
    """

    if IS_CUDA :
        try :
            x_gpu = cp.asarray(x)
            res = cp.exp(x_gpu) / cp.sum(cp.exp(x_gpu))
            return cp.asnumpy(res)
        except :
            pass

    res = np.exp(x) / np.sum(np.exp(x))
    return res



def get(func : str) :
    """Helper function to get an activation function

    Return the appropriate activation function
    depending on the given string

    Args
        func:
            Query string for the requested activation function

    Returns
        Appropriate function
    """
    # TODO : handle wrong inputs
    if not isinstance(func, str) :
        ...

    func = func.strip().lower()

    if func == "sigmoid" :
        return sigmoid
    elif func == "tanh" :
        return tanh
    elif func == "relu" :
        return relu
    elif func == "leakyrelu" :
        return leakyrelu
    # elif func == "softmax" :
    #     return softmax


"""Base class file for callback methods"""

class BaseCallback :
    """Callback methods base class

    CallBack methods are those methods which
    are called at the end of each epochs to
    perform certain tasks.

    This is the base class for such callbacks.
    All callbacks need to inherit from this
    and implement the call method.
    
    """

    def __init__(self) -> None:
        pass

    def run(self) -> None :
        """Method that runs after each epoch

        Every subclass needs to implement this method
        
        """
        pass


"""Shamboflow : a Tensorflow competitor

A state of the art machine learning library with blazing fast speeds and GPU support
"""

# Startup checks and constants here

IS_CUDA = False
"""Constant to decide whether to use GPU for computations."""
try :
    import cupy as cp
    IS_CUDA = cp.cuda.is_available()

    if IS_CUDA :
        print("CUDA enabled GPU found")
        print("-----------------------")

        _cuda_version = cp.cuda.runtime.runtimeGetVersion()
        _driver_version = cp.cuda.runtime.driverGetVersion()

        print(f"CUDA Version: {_cuda_version // 1000}.{(_cuda_version % 1000) // 10}")
        print(f"CUDA Driver Version: {_driver_version // 1000}.{(_driver_version % 1000) // 10}")

        _props = cp.cuda.runtime.getDeviceProperties(0)
        print(f"Name: {_props['name'].decode()}")
        print(f"Compute Capability: {_props['major']}.{_props['minor']}")
        print(f"Total Memory: {_props['totalGlobalMem'] / 1e9:.2f} GB")
        print("\n-----------------------------------------------\n")
except :
    pass

# Add Colorama fixes to terminal
from colorama import just_fix_windows_console
just_fix_windows_console()

from shamboflow import models
from shamboflow import layers
from shamboflow import callbacks


"""Some common models"""

import numpy as np
import cupy as cp

from shamboflow import IS_CUDA
from shamboflow.engine.base_layers import BaseLayer
from shamboflow.engine.base_models import BaseModel
from shamboflow.engine import losses
from shamboflow.engine import d_losses, d_activations

from tqdm import tqdm, trange
from colorama import Fore, Back, Style

class Sequential(BaseModel) :
    """A simple sequential model with multiple layers one after the other
    
    This is the simplest model type
    that is commonly used. It has
    multiple layers in it sequentially
    connected by dense connected directed
    graph between neurons.

    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def add(self, layer: BaseLayer) -> None:
        return super().add(layer)

    def compile(self, loss : str, learning_rate : float = 0.001, verbose : bool = False, **kwargs) -> None:
        """Method to initialize and compile the model
        
        Initializes all the required values
        and performs required operations.

        Args
            loss:
                The loss function to use. Available => ``mean_squared_error``
            learning_rate:
                The learning_rate to use while fitting data
            verbose:
                Should progress be displayed in details

        """

        self.learning_rate = learning_rate
        self.loss_str = loss
        self.loss = losses.get(loss)

        if verbose :
            print(Fore.CYAN + "Compiling Model ...")
            print(Style.RESET_ALL + "Building layers")

        with tqdm(self.layers) as pbar :
            for layer in pbar :
                layer.build()
                self.parameters += layer.size
                if verbose :
                    pbar.update()

        if verbose :
            print(Fore.GREEN + "Finished building layers!")
            print(Style.RESET_ALL + "Generating weight matrices")
            
        for i in trange(len(self.layers) - 1) :
            size_a = self.layers[i].size
            size_b = self.layers[i+1].size

            self.parameters += size_a * size_b

            if IS_CUDA :
                weight_mat = cp.random.uniform(-0.5, 0.5, (size_a, size_b))
                self.weights.append(weight_mat)
            else :
                weight_mat = np.random.uniform(-0.5, 0.5, (size_a, size_b))
                self.weights.append(weight_mat)

        if verbose :
            print(Fore.GREEN + "Finished generating weight matrices")

        print(Fore.CYAN + "Model successfully compiled")
        print(Style.RESET_ALL)

        self.is_compiled = True


    def fit(self, train_x : np.ndarray, train_y : np.ndarray, epochs : int, **kwargs) -> None:
        """Method to train the model and fit the data

        It runs the training where the
        network does the learning.

        Args
            train_x:
                The features of the dataset
            train_y:
                The label of the dataset
            epochs:
                The number of steps to run the training for
            
            
        Keyword Args
            validation_x: validation data set features
            validation_y: validation data set labels
            callbacks:  A list of callback methods
        
        """

        if not self.is_compiled :
            print(Fore.WHITE + Back.RED + "Model not compiled ! Compile the model first.")
            print(Style.RESET_ALL)
            return

        self.train_data_x = train_x
        self.train_data_y = train_y
        self.epochs = epochs

        if 'validation_x' in kwargs and 'validation_y' in kwargs :
            self.has_validation_data = True
            self.validation_x = kwargs['validation_x']
            self.validation_y = kwargs['validation_y']

        if 'callbacks' in kwargs :
            self.callbacks = kwargs['callbacks']

        num_rows = self.train_data_x.shape[0] - 1

        while self.is_fitting and self.current_epoch < self.epochs :

            with tqdm(total=num_rows + 1) as pbar :
                def row_iter(x) :
                    num_layer = -1

                    # Forward propagation
                    for layer in self.layers :
                        num_layer += 1
                        if num_layer == 0 :
                            layer.compute(input = x)
                            continue

                        if IS_CUDA :
                            op_gpu = cp.asarray(self.layers[num_layer - 1].output_array)
                            weight_gpu = self.weights[num_layer - 1]
                            layer.compute(cp.asnumpy(cp.matmul(op_gpu, weight_gpu)))
                        else :
                            op = self.layers[num_layer - 1].output_array
                            weight = self.weights[num_layer - 1]
                            layer.compute(np.matmul(op, weight))
                    
                    self.error_val = self.loss(self.layers[num_layer].output_array, self.train_data_y[num_rows])
                    acc = np.subtract(self.train_data_y[num_rows], self.layers[num_layer].output_array)
                    self.accuracy_val = (self.layers[num_layer].size - np.count_nonzero(acc)) / self.layers[num_layer].size

                    self.metrics['acc'] = self.accuracy_val
                    self.metrics['loss'] = self.error_val

                    # Validation
                    if self.has_validation_data :
                        self.evaluate(self.validation_x, self.validation_y, is_validation=True)

                    # BackPropagation
                    d_loss_fun = d_losses.get(self.loss_str)
                    d_act_fun = d_activations.get(self.layers[num_layer].activation_str)

                    if IS_CUDA :
                        ## Compute Gradients for output layer
                        d_loss_res_gpu = cp.asarray(d_loss_fun(self.layers[num_layer].output_array, self.train_data_y[num_rows]))
                        d_act_res_gpu = cp.asarray(d_act_fun(self.layers[num_layer].midway, leakyrelu_slope=self.layers[num_layer].leakyrelu_slope))
                        gradient_op = cp.multiply(d_loss_res_gpu, d_act_res_gpu)
                        self.layers[num_layer].error_array = gradient_op
                        weight_gradient = cp.multiply.outer(gradient_op, self.layers[num_layer - 1].output_array)
                        self.weights[num_layer - 1] = cp.add(self.weights[num_layer - 1], cp.multiply(self.learning_rate, weight_gradient).T)

                        ## Compute Gradient for output layer bias
                        # gradient_op is the bias gradient itself
                        # Adjust bias
                        self.layers[num_layer].bias_array = cp.add(self.layers[num_layer].bias_array, cp.multiply(self.learning_rate, gradient_op))

                        # Hidden layers
                        for i in range(len(self.layers) - 2, 0, -1) :
                            d_act_fun = d_activations.get(self.layers[i].activation_str)
                            error_next = self.layers[i+1].error_array
                            d_act_res_hidden = cp.asarray(d_act_fun(self.layers[i].midway, leakyrelu_slope=self.layers[i].leakyrelu_slope))
                            hidden_weight = cp.transpose(self.weights[i])
                            hidden_err_midway = cp.matmul(error_next, hidden_weight)
                            hidden_error = cp.multiply(hidden_err_midway, d_act_res_hidden)
                            self.layers[i].error_array = hidden_error

                            weight_gradient_hidden = cp.multiply.outer(hidden_error, self.layers[i-1].output_array)
                            self.weights[i-1] = cp.add(self.weights[i-1], cp.multiply(self.learning_rate, weight_gradient_hidden).T)

                            # Bias
                            self.layers[i].bias_array = cp.add(self.layers[i].bias_array, cp.multiply(self.learning_rate, hidden_error))
                        
                    else :
                        ## Compute Gradients for output layer
                        d_loss_res = d_loss_fun(self.layers[num_layer].output_array, self.train_data_y[num_rows])
                        d_act_res = d_act_fun(self.layers[num_layer].midway, leakyrelu_slope=self.layers[num_layer].leakyrelu_slope)
                        gradient_op = np.multiply(d_loss_res, d_act_res)
                        self.layers[num_layer].error_array = gradient_op
                        weight_gradient = np.multiply.outer(gradient_op, self.layers[num_layer - 1].output_array)
                        self.weights[num_layer - 1] = np.add(self.weights[num_layer - 1], np.multiply(self.learning_rate, weight_gradient).T)

                        ## Compute Gradient for output layer bias
                        # gradient_op is the bias gradient itself
                        # Adjust bias
                        self.layers[num_layer].bias_array = np.add(self.layers[num_layer].bias_array, np.multiply(self.learning_rate, gradient_op))

                        # Hidden layers
                        for i in range(len(self.layers) - 2, 0, -1) :
                            d_act_fun = d_activations.get(self.layers[i].activation_str)
                            error_next = self.layers[i+1].error_array
                            d_act_res_hidden = d_act_fun(self.layers[i].midway, leakyrelu_slope=self.layers[i].leakyrelu_slope)
                            hidden_weight = self.weights[i]
                            hidden_err_midway = np.matmul(error_next, hidden_weight)
                            hidden_error = np.multiply(hidden_err_midway, d_act_res_hidden)
                            self.layers[i].error_array = hidden_error

                            weight_gradient_hidden = np.multiply.outer(hidden_error, self.layers[i-1].output_array)
                            self.weights[i-1] = np.add(self.weights[i-1], np.multiply(self.learning_rate, weight_gradient_hidden).T)

                            # Bias
                            self.layers[i].bias_array = np.add(self.layers[i].bias_array, np.multiply(self.learning_rate, hidden_error))

                    pbar.set_description(f"Epoch: {self.current_epoch}")
                    pbar.set_postfix_str(f"Accuracy: {self.accuracy_val}, Loss: {self.error_val}")
                    pbar.update(1)    

                np.apply_along_axis(row_iter, 1, self.train_data_x)
            
            # Call the callback methods
            for callback in self.callbacks :
                callback.run(model=self)

            self.current_epoch += 1


    def evaluate(self, x_data : np.ndarray, y_data : np.ndarray, **kwargs) -> None:
        """Method to evaluate the model with test data
        
        Args
            x_data:
                The features of the dataset
            y_data:
                The label of the dataset
            
        Keyword Args
            is_validation: Whether the model is called during validation set checking step of training
        """

        is_val = False
        if 'is_validation' in kwargs :
            is_val = kwargs['is_validation']

        num_rows = x_data.shape[0]
        test_error_val = 0.0
        test_accuracy_val = 0.0

        with tqdm(total=num_rows) as pbar :
            def row_iter(x) :
                num_layer = -1
                global test_error_val
                global test_accuracy_val

                for layer in self.layers :
                    num_layer += 1
                    if num_layer == 0 :
                        layer.compute(input = x)
                        continue

                    if IS_CUDA :
                        op_gpu = cp.asarray(self.layers[num_layer - 1].output_array)
                        weight_gpu = self.weights[num_layer - 1]
                        layer.compute(cp.matmul(op_gpu, weight_gpu))
                    else :
                        op = self.layers[num_layer - 1].output_array
                        weight = self.weights[num_layer - 1]
                        layer.compute(np.matmul(op, weight))
                    
                test_error_val = self.loss(self.layers[num_layer].output_array, y_data[num_rows])
                acc = np.subtract(y_data[num_rows], self.layers[num_layer].output_array)
                test_accuracy_val = (self.layers[num_layer].size - np.count_nonzero(acc)) / self.layers[num_layer].size

                if is_val :
                    self.metrics['val_loss'] = test_error_val
                    self.metrics['val_acc'] = test_accuracy_val
                    return

                pbar.set_postfix_str(f"Accuracy: {test_accuracy_val}, Loss: {test_error_val}")
                pbar.update(1)

            np.apply_along_axis(row_iter, 1, x_data)

        print(f"Accuracy: {test_accuracy_val}, Error: {test_error_val}")


    def summary(self) -> None:
        """Prints a summary of the model once compiled"""

        if not self.is_compiled :
            print(Back.RED + Fore.WHITE + "Model has not been compiled.\nCompile the model first using model.compile()")
            print(Style.RESET_ALL)
            return

        print(Fore.WHITE + "Model type : " + Fore.CYAN + "Sequential\n")
        print(Fore.WHITE + "Layers: ")

        for layer in self.layers :
            print("-> " + Fore.CYAN + layer.name + Fore.WHITE + f" Neurons: {layer.size} Activation: {layer.activation_str if layer.trainable else 'Not trainable'} Trainable: {layer.trainable}")

        print(Style.RESET_ALL)
        print(f"\nTrainable Params: {self.parameters}")


    def save(self, save_path : str) -> None:
        """Method to save the model to disk
        
        Args
            save_path:
                Path to where the model will be saved, along with the name of the model file
        """

        import pickle

        with open(save_path, 'wb') as f :
            pickle.dump(self, f)
            print(f"Saved model to: {save_path}")

    def predict(self, input_x : np.ndarray) -> np.ndarray :
        """Method to predict data labels
        
        Args
            input_x:
                The features of the data to predict

        Returns
            The final output layer with the generated values
        """
        
        num_layer = -1
        pred_layers = self.layers

        for layer in pred_layers :
            num_layer += 1
            if num_layer == 0 :
                layer.compute(input = input_x)
                continue

            if IS_CUDA :
                op_gpu = cp.asarray(pred_layers[num_layer - 1].output_array)
                weight_gpu = self.weights[num_layer - 1]
                layer.compute(cp.matmul(op_gpu, weight_gpu))
            else :
                op = pred_layers[num_layer - 1].output_array
                weight = self.weights[num_layer - 1]
                layer.compute(np.matmul(op, weight))
        
        return pred_layers[num_layer].output_array
        

def load_model(path_to_model : str) -> BaseModel :
    """Method to load a model from disk
    
    Args
        path_to_model:
            path to the model file
    
    Returns
        The model object

    """

    import pickle

    with open(path_to_model, 'rb') as f :
        model = pickle.load(f)
        return model


"""Partial derivatives of loss functions"""

import numpy as np
import cupy as cp

from shamboflow import IS_CUDA

def d_MSE(pred_val : np.ndarray, obs_val : np.ndarray) -> np.ndarray :
    """Derivative of Mean Squared Error loss function
    
    ``Formula : y_pbserved - y_predicted``
    """

    if IS_CUDA :
        pred_val = cp.asarray(pred_val)
        obs_val = cp.asarray(obs_val)
        res = cp.subtract(obs_val, pred_val)
        return cp.asnumpy(res)
    
    return np.subtract(obs_val, pred_val)

def get(func : str) :
    """Helper function to get a loss function derivative

    Return the appropriate loss function derivative
    depending on the given string

    Args
        func:
            Query string for the requested loss function derivative

    Returns
        Appropriate function
    """

    func = func.strip().lower()

    if func == "mean_squared_error" :
        return d_MSE


"""Base class for layers"""

import numpy as np

class BaseLayer :
    """The base class for layers to inherit from.

    A layer is an object that is a part of
    the neural network. It performs calculations
    on neurons or neurons present in it
    commonly. It can also not have computable
    neurons, instead can also be used to filter
    data from previous layers.

    Attributes
        name:
            The name of the layer
        trainable:
            Is the layer trainable
        is_built:
            Is the layer built and initialized
    """

    def __init__(self, name : str, trainable : bool = True) -> None:
        """Constructor for the base layer class

        Args
            name:
                Name of the layer
            trainable:
                Is the layer trainable. Default is True
        
        """
        self.name : str = name
        """Name of the layer"""
        self.trainable : bool = trainable
        """Is the layer trainable"""
        self.is_built : bool = False
        """has the layer been built and compiled"""

        self.bias_array : np.ndarray = None
        """Bias values of the layer"""
        self.output_array : np.ndarray = None
        """Output matrix of the layer"""
        self.midway : np.ndarray = None
        """Output value before applying avtivation function"""
        self.error_array : np.ndarray = None
        """Error values of the layer"""
        self.leakyrelu_slope : float = 0.0
        """Slope value if using leakyrelu activation function"""

        self.activation = None
        """Activation function used by the layer"""
        self.activation_str : str = None
        """The name of the activation function"""

    def build(self) -> None :
        """Method to build and initialize the layer.

        This method is to be overidden by child
        classes to build them by their own logic.
        
        """
        self.is_built = True

    def compute(self) -> None :
        """Method to compute various parameter
        
        This method is to be implemented in inherited classes
        """
        pass


"""A collection of common callback functions"""

from colorama import Fore, Back, Style

from shamboflow.engine.base_callback import BaseCallback
from shamboflow.engine.base_models import BaseModel

class EarlyStopping(BaseCallback) :
    """Early Stopper

    This callback method monitors a given
    metric and then stops the training
    early if the metric doesn't improve
    for a given amount of time
    
    """

    def __init__(self, monitor : str = 'loss', patience : int = 10, verbose : bool = False, **kwargs) -> None :
        """Initialize

        Args
            monitor:
                The metric to monitor. It is one of the 4: ``loss``, ``acc``, ``val_loss``, ``val_acc``
            patience:
                How many epoch to monitor before stopping
            verbose:
                Log callback function logs
        
        """

        self._monitor = monitor
        self._patience = patience
        self._verbose = verbose

        self._metric_old = 0.0
        self._patience_ctr = 0

    def run(self, model : BaseModel) -> None:
        """The callback method that will be called after each epoch
        
        Args
            model: The model on which callback wil be called
        """

        if self._monitor in ('val_loss', 'val_acc') :
            if not model.has_validation_data :
                self._monitor = self._monitor.removeprefix('val_')

        current_metric = model.metrics[self._monitor]
        if model.current_epoch == 0 :
            self._metric_old = current_metric
            return
        
        if 'loss' in self._monitor :
            if current_metric >= self._metric_old :
                self._patience_ctr += 1

                if self._verbose :
                    print(f"[EarlyStopping] : {self._monitor} has not improved for the past {self._patience_ctr} epochs.")

                if self._patience_ctr == self._patience :
                    print(Back.RED + Fore.WHITE + f"Early Stopping, {self._monitor} has not improved for past {self._patience} epochs")
                    print(Style.RESET_ALL)
                    model.stop()
            else :
                self._patience_ctr = 0
        else :
            if current_metric <= self._metric_old :
                self._patience_ctr += 1

                if self._verbose :
                    print(f"[EarlyStopping] : {self._monitor} has not improved for the past {self._patience_ctr} epochs.")
                    
                if self._patience_ctr == self._patience :
                    print(Back.RED + Fore.WHITE + f"Early Stopping, {self._monitor} has not improved for past {self._patience} epochs")
                    print(Style.RESET_ALL)
                    model.stop()
            else :
                self._patience_ctr = 0

        self._metric_old = current_metric


class ReduceLROnPlateau(BaseCallback) :
    """Reduce Learning Rate on Plateau callback
    
    This callback function monitors a
    given metric for a set amount of
    iterations and reduces the learning
    rate if the metric doesn't improve
    for the given duration.
    """

    def __init__(self, monitor : str = 'loss', patience : int = 10, factor : float = 0.9, min_val : float = 1e-10, verbose : bool = False, **kwargs) -> None :
        """Initialize
        
        Args
            monitor:
                The metric to monitor. It is one of the 4: ``loss``, ``acc``, ``val_loss``, ``val_acc``
            patience:
                How many epoch to monitor before stopping
            factor:
                fraction to which the learning rate will be lowered to. Note - This is not by how much to reduce but how much to reduce to
            min_val:
                Min possible learning rate
            verbose:
                Log callback function logs
        """

        self._monitor = monitor
        self._patience = patience
        self._factor = factor
        self._min_val = min_val
        self._verbose = verbose

        self._metric_old = 0.0
        self._patience_ctr = 0

    def run(self, model : BaseModel) -> None:
        """The callback method that will be called after each epoch
        
        Args
            model: The model on which callback wil be called
        """

        if self._monitor in ('val_loss', 'val_acc') :
            if not model.has_validation_data :
                self._monitor = self._monitor.removeprefix('val_')

        current_metric = model.metrics[self._monitor]
        if model.current_epoch == 0 :
            self._metric_old = current_metric
            return
        
        if 'loss' in self._monitor :
            if current_metric >= self._metric_old :
                self._patience_ctr += 1

                if self._verbose :
                    print(f"[ReduceLROnPlateau] : {self._monitor} has not improved for the past {self._patience_ctr} epochs.")

                if self._patience_ctr == self._patience :
                    new_LR = self._factor * model.learning_rate
                    model.learning_rate = new_LR if new_LR > self._min_val else model.learning_rate

                    print(Back.RED + Fore.WHITE + f"Reducing Learning Rate, {self._monitor} has not improved for past {self._patience} epochs. New Learning rate is {model.learning_rate}")
                    print(Style.RESET_ALL)
            else :
                self._patience_ctr = 0
        else :
            if current_metric <= self._metric_old :
                self._patience_ctr += 1

                if self._verbose :
                    print(f"[ReduceLROnPlateau] : {self._monitor} has not improved for the past {self._patience_ctr} epochs.")
                    
                if self._patience_ctr == self._patience :
                    new_LR = self._factor * model.learning_rate
                    model.learning_rate = new_LR if new_LR > self._min_val else model.learning_rate

                    print(Back.RED + Fore.WHITE + f"Reducing Learning Rate, {self._monitor} has not improved for past {self._patience} epochs. New Learning rate is {model.learning_rate}")
                    print(Style.RESET_ALL)
            else :
                self._patience_ctr = 0

        self._metric_old = current_metric

class ModelCheckpoint(BaseCallback) :
    """Model Checkpointing
    
    This callback function monitors a
    metric and saves the instance of model
    to disk at every epoch or on a condition
    """

    def __init__(self, save_path : str = "model.ckpt", monitor : str = 'loss', save_best_only : bool = True, verbose : bool = False, **kwargs) -> None :
        """Initialize
        
        Args
            save_path:
                Save path for the model
            monitor:
                The metric to monitor. It is one of the 4: ``loss``, ``acc``, ``val_loss``, ``val_acc``
            save_best_only:
                whether to save only the best model according to the monitor
            verbose:
                Log callback function logs

        """

        self._save_path = save_path
        self._monitor = monitor
        self._save_best = save_best_only
        self._verbose = verbose

        self._metric_old = 0.0

    def run(self, model : BaseModel) -> None:
        """The callback method that will be called after each epoch
        
        Args
            model: The model on which callback wil be called
        """

        if not self._save_best :
            if self._verbose :
                print(f"[ModelCheckpoint] : Saving to {self._save_path}")
            model.save(self._save_path)
            return

        if self._monitor in ('val_loss', 'val_acc') :
            if not model.has_validation_data :
                self._monitor = self._monitor.removeprefix('val_')

        current_metric = model.metrics[self._monitor]
        if model.current_epoch == 0 :
            self._metric_old = current_metric
            return
        
        if 'loss' in self._monitor :
            if current_metric < self._metric_old :
                if self._verbose :
                    print(f"[ModelCheckpoint] : Metric has improved. Saving to {self._save_path}")
                model.save(self._save_path)
        else :
            if current_metric > self._metric_old :
                if self._verbose :
                    print(f"[ModelCheckpoint] : Metric has improved. Saving to {self._save_path}")
                model.save(self._save_path)

        self._metric_old = current_metric


import unittest
import numpy as np
from shamboflow.engine.activations import signmoid, relu, tanh, leakyrelu, softmax

class TestActivation(unittest.TestCase) :

    def test_relu(self) :
        x = np.array([-2, 3, 4, -4])
        x_res = np.array([0, 3, 4, 0])

        np.testing.assert_array_equal(relu(x), x_res)

    def test_sigmoid(self) :
        x = np.array([-2, 3, 4, -4])
        x_res = np.array([0.119203, 0.952574, 0.982014, 0.017986])

        np.testing.assert_array_almost_equal(signmoid(x), x_res)

unittest.main()


"""Several common layers used in Neural Networks"""

import numpy as np
import cupy as cp

from shamboflow import IS_CUDA
from shamboflow.engine.base_layers import BaseLayer
from shamboflow.engine.activations import get

class Dense(BaseLayer) :
    """A Simple 1D layer

    A Dense layer is a simple 1D layer
    that just has a given number of
    neurons. Its the most common
    and basic layer.

    Attributes
        size: The number of neurons in the layer
        bias: An array of bias values for a neuron
        activation: The activation function to apply to this layer
        output: An array of output values after applying activation function
    
    """

    def __init__(self, size : int, activation : str = None, **kwargs) -> None:
        """Constructor for Dense Layer

        Args
            size:
                The number of neurons in the layer
            activation:
                The activation function to use for the layer. Available => `sigmoid`, `tanh`, `relu`, `leakyrelu`

        Keyword Args
            leakyrelu_slope:
                The slope value if leakyrelu is used as the activation function
        """
        super().__init__("Dense", activation != None)

        self.size : int = size
        """Size of the layer"""
        if activation != None :
            self.activation = get(activation)
            self.activation_str = activation

        if "leakyrelu_slope" in kwargs : 
            self.leakyrelu_slope = kwargs.get("leakyrelu_slope")
        

    def build(self) -> None:
        """Overidden Build method

        This method initializes the bias and output data array.
        """
        if IS_CUDA :
            self.bias_array = cp.random.uniform(-0.5, 0.5, self.size)
            self.output_array = cp.zeros(self.size)
        else :
            self.bias_array = np.random.uniform(-0.5, 0.5, self.size)
            self.output_array = np.zeros(self.size)

        super().build()
    
    def compute(self, input : np.ndarray) -> np.ndarray :
        """Method to perform computation on data

        This method accepts an input vector
        that is the output vector of the
        previous layer in the network. Then
        output values of this layer is calculated.

        The input values are simply added with the
        bias and then passed through the activation
        function.

        Args
            input:
                The input vector
        
        Returns
            The output vector after computaion
        
        """

        if not self.trainable :
            self.output_array = input
            return self.output_array

        if IS_CUDA :
            input_gpu = cp.asarray(input)
            self.midway = cp.add(input_gpu, self.bias_array)
            self.output_array = self.activation(cp.asnumpy(self.midway), leakyrelu_slope=self.leakyrelu_slope)
            return self.output_array

        self.midway = np.add(input, self.bias_array)
        self.output_array = self.activation(self.midway, leakyrelu_slope=self.leakyrelu_slope)
        return self.output_array


"""Base class for models"""

import numpy as np

from shamboflow.engine.base_layers import BaseLayer

class BaseModel :
    """The Base class for models
    
    All other models need to
    inherit from this class.
    It provides a template with
    all the methods a model will
    need to execute.

    Attributes
        layers:
            A list of layers present in the model in the order from input to output
        weights:
            A list of all weight matrices for the layers in the model
        loss:
            The loss/cost function to use calculate error of the model at a given state
        train_data_x:
            The training data or features
        train_data_y:
            Training features' corresponding labels

    And more attributes

    """

    def __init__(self, **kwargs) -> None:
        """Initialize the required attributes.

        Also optionally takes in a list of
        layers as a parameter.

        Keyword Args
            layers: It takes a list of layers and them to the model
        
        """

        self.layers : list[BaseLayer] = []
        """A list of layers in the model"""
        self.weights : list[np.ndarray] = []
        """A list of all weights between each layers"""
        self.loss = None
        """The loss function used by the model"""
        self.loss_str : str = ""
        """The name of the loss function used"""
        self.learning_rate : float = 0.0
        """The learning rate of the model"""
        self.train_data_x : np.ndarray = None
        """The feature list of the dataset"""
        self.train_data_y : np.ndarray = None
        """The labels for each data in the dataset"""
        self.validation_x : np.ndarray = None
        """Features of validation set"""
        self.validation_y : np.ndarray = None
        """Labels of the validation set"""
        self.has_validation_data : bool = False
        """Boolean flag fo if the the model has validation data"""
        self.test_data_x : np.ndarray = None
        """Features of the test dataset"""
        self.test_data_y : np.ndarray = None
        """Labels for the test dataset"""
        self.epochs : int = None
        """Number of epochs to train for"""
        self.callbacks : list = []
        """A list of callback functions"""

        self.error_val : float = 0.0
        """The loss value of the model at any instance"""
        self.accuracy_val : float = 0.0
        """The accuracy value of the model at any instance"""

        self.metrics : dict = {'loss' : 0.0, 'acc' : 0.0, 'val_loss' : 0.0, 'val_acc' : 0.0}
        """A dictionary of metrics for the model. Includes ``loss``, ``acc``, ``val_loss``, ``val_acc``"""
        self.current_epoch : int = 0
        """The current epoch of the model during training"""
        self.is_fitting : bool = True
        """Is the model training right now"""

        self.parameters : int = 0
        """Amount of trainable parameters in the model"""

        self.is_compiled : bool = False
        """Has the model been compiled"""

        if "layers" in kwargs :
            layers_arg = kwargs.get("layers")
            for layer in layers_arg :
                self.add(layer)


    def add(self, layer : BaseLayer) -> None :
        """Method to add layers to the model

        Args
            layer:
                layer to add to the model
        
        """
        self.layers.append(layer)

    def compile(self) -> None :
        """Model compilation method
        
        To be implemented in child
        """
        pass

    def fit(self) -> None :
        """Method to train the model on data
        
        To be implemented in child
        """
        pass

    def stop(self) -> None :
        """Method to stop the training"""
        self.is_fitting = False

    def summary(self) -> None :
        """Prints a summary of the model with all necessary details"""
        pass

    def save(self) -> None :
        """Saves the model to the disk"""
        pass
    
    def evaluate(self) -> None :
        """Evaluates metrics based on a Test dataset"""
        pass
    
    def predict(self) -> None :
        """Calculate inference using the trained model"""
        pass